@@ -2,8 +2,8 @@ package Bifcode;
 use 5.010;
 use strict;
 use warnings;
-use boolean;
-use Carp;
+use boolean ();
+use Carp 'croak';
 use Exporter::Tidy all => [
     qw( encode_bifcode
       decode_bifcode
@@ -13,10 +13,53 @@ use Exporter::Tidy all => [
 
 # ABSTRACT: Serialisation similar to Bencode + undef/UTF8
 
-our $VERSION = '0.001_10';
-our ( $DEBUG, $max_depth );
+our $VERSION = '0.001_11';
+our $max_depth;
 
-sub _msg { sprintf "@_", pos() || 0 }
+sub _error {
+    my $type = shift // croak 'usage: _error($TYPE, [$msg])';
+    my %messages = (
+        Decode             => 'garbage at',
+        DecodeBytes        => 'malformed BYTES length at',
+        DecodeBytesTrunc   => 'unexpected BYTES end of data at',
+        DecodeDepth        => 'nesting depth exceeded at',
+        DecodeTrunc        => 'unexpected end of data at',
+        DecodeFloat        => 'malformed FLOAT data at',
+        DecodeFloatTrunc   => 'unexpected FLOAT end of data at',
+        DecodeInteger      => 'malformed INTEGER data at',
+        DecodeIntegerTrunc => 'unexpected INTEGER end of data at',
+        DecodeTrailing     => 'trailing garbage at',
+        DecodeUTF8         => 'malformed UTF8 string length at',
+        DecodeUTF8Trunc    => 'unexpected UTF8 end of data at',
+        DecodeUsage        => undef,
+        DiffUsage          => 'usage: diff_bifcode($b1, $b2, [$diff_args])',
+        EncodeBytesUndef   => 'Bifcode::BYTES ref is undefined',
+        EncodeFloat        => undef,
+        EncodeFloatUndef   => 'Bifcode::FLOAT ref is undefined',
+        EncodeInteger      => undef,
+        EncodeIntegerUndef => 'Bifcode::INTEGER ref is undefined',
+        DecodeKeyType      => 'dict key is not BYTES or UTF8 at',
+        DecodeKeyDuplicate => 'duplicate dict key at',
+        DecodeKeyOrder     => 'dict key not in sort order at',
+        DecodeKeyValue     => 'dict key is missing value at',
+        EncodeUTF8Undef    => 'Bifcode::UTF8 ref is undefined',
+        EncodeUnhandled    => undef,
+        EncodeUsage        => 'usage: encode_bifcode($arg)',
+        ForceUsage         => 'ref and type must be defined',
+    );
+
+    my $msg = shift // $messages{$type} // '(no message)';
+    $msg =~ s! at$!' at '. ( pos() // 0 )!e;
+
+    eval qq{
+        package Bifcode::Error::$type;
+        use overload
+          bool => sub { 1 },
+          '""' => sub { \${\$_[0]} . ' (' . ( ref \$_[0] ) . ')' },
+          fallback => 1;
+    };
+    bless \$msg, 'Bifcode::Error::' . $type;
+}
 
 my $match = qr/ \G (?|
       (~)
@@ -24,22 +67,18 @@ my $match = qr/ \G (?|
     | (1)
     | (B|U) (?: ( 0 | [1-9] [0-9]* ) : )? 
     | (I) (?: ( 0 | -? [1-9] [0-9]* ) , )?
-    | (F) (?: (-)? ( 0 | [1-9] [0-9]* )
+    | (F) (?: ( (-)? ( 0 | [1-9] [0-9]* )
         \. ( 0 | [0-9]* [1-9] )
-        e (( 0 | -? [1-9] ) [0-9]*) , )?
+        e (( 0 | -? [1-9] ) [0-9]*) ) , )?
     | (\[)
     | (\{)
 ) /x;
 
 sub _decode_bifcode_chunk {
-    warn _msg 'decoding at %s' if $DEBUG;
-
     local $max_depth = $max_depth - 1 if defined $max_depth;
 
-    if ( !m/$match/gc ) {
-        croak _msg m/ \G \z /xgc
-          ? 'unexpected end of data at %s'
-          : 'garbage at %s';
+    unless (m/$match/gc) {
+        croak _error m/ \G \z /xgc ? 'DecodeTrunc' : 'Decode';
     }
 
     if ( $1 eq '~' ) {
@@ -52,99 +91,65 @@ sub _decode_bifcode_chunk {
         return boolean::true;
     }
     elsif ( $1 eq 'B' ) {
-        my $len = $2 // croak _msg 'malformed string length at %s';
-
-        croak _msg 'unexpected end of string data starting at %s'
-          if $len > length() - pos();
+        my $len = $2 // croak _error 'DecodeBytes';
+        croak _error 'DecodeBytesTrunc' if $len > length() - pos();
 
         my $data = substr $_, pos(), $len;
         pos() = pos() + $len;
 
-        warn _msg BYTES => "(length $len) at %s", if $DEBUG;
         return $data;
     }
     elsif ( $1 eq 'U' ) {
-        my $len = $2 // croak _msg 'malformed string length at %s';
-
-        croak _msg 'unexpected end of string data starting at %s'
-          if $len > length() - pos();
+        my $len = $2 // croak _error 'DecodeUTF8';
+        croak _error 'DecodeUTF8Trunc' if $len > length() - pos();
 
         utf8::decode( my $str = substr $_, pos(), $len );
         pos() = pos() + $len;
-
-        warn _msg
-          UTF8 => "(length $len)",
-          $len < 200 ? "[$str]" : (), 'at %s'
-          if $DEBUG;
-
         return $str;
     }
     elsif ( $1 eq 'I' ) {
-        if ( defined $2 ) {
-            warn _msg INTEGER => $2, 'at %s' if $DEBUG;
-            return $2;
-        }
-        croak _msg 'unexpected end of data at %s' if m/ \G \z /xgc;
-        croak _msg 'malformed integer data at %s';
+        return $2 if defined $2;
+        croak _error 'DecodeIntegerTrunc' if m/ \G \z /xgc;
+        croak _error 'DecodeInteger';
     }
     elsif ( $1 eq 'F' ) {
-        croak _msg 'malformed float data at %s'
-          if $3 eq '0'
-          and $4 eq '0'
-          and ( $2 or $5 ne '0' );
+        if ( !defined $2 ) {
+            ;
+            croak _error 'DecodeFloatTrunc' if m/ \G \z /xgc;
+            croak _error 'DecodeFloat';
+        }
+        croak _error 'DecodeFloat'
+          if $4 eq '0'     # mantissa A.b
+          and $5 eq '0'    # mantissa a.B
+          and ( $3 or $6 ne '0' );    # sign or exponent
 
-        warn _msg
-          FLOAT => ( $2 // '' ) . $3 . '.' . $4 . 'e' . $5,
-          'at %s'
-          if $DEBUG;
-        return ( $2 // '' ) . $3 . '.' . $4 . 'e' . $5;
+        return $2;
     }
     elsif ( $1 eq '[' ) {
-        warn _msg 'LIST at %s' if $DEBUG;
-
-        croak _msg 'nesting depth exceeded at %s'
-          if defined $max_depth and $max_depth < 0;
+        croak _error 'DecodeDepth' if defined $max_depth and $max_depth < 0;
 
         my @list;
         until (m/ \G \] /xgc) {
-            warn _msg 'list not terminated at %s, looking for another element'
-              if $DEBUG;
             push @list, _decode_bifcode_chunk();
         }
         return \@list;
     }
     elsif ( $1 eq '{' ) {
-        warn _msg 'DICT at %s' if $DEBUG;
-
-        croak _msg 'nesting depth exceeded at %s'
-          if defined $max_depth and $max_depth < 0;
+        croak _error 'DecodeDepth' if defined $max_depth and $max_depth < 0;
 
         my $last_key;
         my %hash;
         until (m/ \G \} /xgc) {
-            warn _msg 'dict not terminated at %s, looking for another pair'
-              if $DEBUG;
+            croak _error 'DecodeTrunc' if m/ \G \z /xgc;
+            croak _error 'DecodeKeyType' unless m/ \G (B|U) /xgc;
 
-            croak _msg 'unexpected end of data at %s'
-              if m/ \G \z /xgc;
+            pos() = pos() - 1;
+            my $key = _decode_bifcode_chunk();
 
-            my $key;
-            if (m/ \G (B|U) /xgc) {
-                pos() = pos() - 1;
-                $key = _decode_bifcode_chunk();
-            }
-            else {
-                croak _msg 'dict key is not a string at %s';
-            }
-
-            croak _msg 'duplicate dict key at %s'
-              if exists $hash{$key};
-
-            croak _msg 'dict key not in sort order at %s'
+            croak _error 'DecodeKeyDuplicate' if exists $hash{$key};
+            croak _error 'DecodeKeyOrder'
               if defined $last_key and $key lt $last_key;
-
-            croak _msg 'dict key is missing value at %s'
-              if m/ \G \} /xgc;
+            croak _error 'DecodeKeyValue' if m/ \G \} /xgc;
 
             $last_key = $key;
             $hash{$key} = _decode_bifcode_chunk();
@@ -156,11 +161,13 @@ sub _decode_bifcode_chunk {
 sub decode_bifcode {
     local $_         = shift;
     local $max_depth = shift;
-    croak 'decode_bifcode: too many arguments: ' . "@_" if @_;
-    croak 'decode_bifcode: only accepts bytes' if utf8::is_utf8($_);
+
+    croak _error 'DecodeUsage', 'decode_bifcode: too many arguments' if @_;
+    croak _error 'DecodeUsage', 'decode_bifcode: only accepts bytes'
+      if utf8::is_utf8($_);
 
     my $deserialised_data = _decode_bifcode_chunk();
-    croak _msg 'trailing garbage at %s' if $_ !~ m/ \G \z /xgc;
+    croak _error 'DecodeTrailing' if $_ !~ m/ \G \z /xgc;
     return $deserialised_data;
 }
 
@@ -192,15 +199,6 @@ sub _encode_bifcode {
                 'U' . length($str) . ':' . $str;
             }
         }
-        elsif ( $type eq 'SCALAR' or $type eq 'Bifcode::BYTES' ) {
-            croak 'Bifcode::BYTES must be defined' unless defined $$_;
-            'B' . length($$_) . ':' . $$_;
-        }
-        elsif ( $type eq 'Bifcode::UTF8' ) {
-            my $str = $$_ // croak 'Bifcode::UTF8 must be defined';
-            utf8::encode($str);    #, sub { croak 'invalid Bifcode::UTF8' } );
-            'U' . length($str) . ':' . $str;
-        }
         elsif ( $type eq 'ARRAY' ) {
             '[' . join( '', map _encode_bifcode($_), @$_ ) . ']';
         }
@@ -224,46 +222,50 @@ sub _encode_bifcode {
                   }
             ) . '}';
         }
+        elsif ( $type eq 'SCALAR' or $type eq 'Bifcode::BYTES' ) {
+            $$_ // croak _error 'EncodeBytesUndef';
+            'B' . length($$_) . ':' . $$_;
+        }
         elsif ( $type eq 'boolean' ) {
             $$_ ? '1' : '0';
         }
         elsif ( $type eq 'Bifcode::INTEGER' ) {
-            croak 'Bifcode::INTEGER must be defined' unless defined $$_;
-            if ( $$_ =~ m/\A (?: 0 | -? [1-9] [0-9]* ) \z/x ) {
-                sprintf 'I%s,', $$_;
-            }
-            else {
-                croak 'invalid integer: ' . $$_;
-            }
+            $$_ // croak _error 'EncodeIntegerUndef';
+            croak _error 'EncodeInteger', 'invalid integer: ' . $$_
+              unless $$_ =~ m/\A (?: 0 | -? [1-9] [0-9]* ) \z/x;
+            sprintf 'I%s,', $$_;
         }
         elsif ( $type eq 'Bifcode::FLOAT' ) {
-            croak 'Bifcode::FLOAT must be defined' unless defined $$_;
-            if ( $$_ =~ $number_qr ) {
-                my $x = 'F' . ( 0 + $1 )    # remove leading zeros
-                  . '.' . ( $3 // 0 ) . 'e' . ( 0 + ( $5 // 0 ) ) . ',';
-                $x =~ s/ ([1-9]) (0+ e)/.${1}e/x;    # remove trailing zeros
-                $x;
-            }
-            else {
-                croak 'invalid float: ' . $$_;
-            }
+            $$_ // croak _error 'EncodeFloatUndef';
+            croak _error 'EncodeFloat', 'invalid float: ' . $$_
+              unless $$_ =~ $number_qr;
+
+            my $x = 'F' . ( 0 + $1 )    # remove leading zeros
+              . '.' . ( $3 // 0 ) . 'e' . ( 0 + ( $5 // 0 ) ) . ',';
+            $x =~ s/ ([1-9]) (0+ e)/.${1}e/x;    # remove trailing zeros
+            $x;
+        }
+        elsif ( $type eq 'Bifcode::UTF8' ) {
+            my $str = $$_ // croak _error 'EncodeUTF8Undef';
+            utf8::encode($str);    #, sub { croak 'invalid Bifcode::UTF8' } );
+            'U' . length($str) . ':' . $str;
         }
         else {
-            croak 'unhandled data type: ' . $type;
+            croak _error 'EncodeUnhandled', 'unhandled data type: ' . $type;
         }
     } @_;
 }
 
 sub encode_bifcode {
-    croak 'usage: encode_bifcode($arg)' if @_ != 1;
-    goto &_encode_bifcode;
+    croak _error 'EncodeUsage' if @_ != 1;
+    (&_encode_bifcode)[0];
 }
 
 sub force_bifcode {
     my $ref  = shift;
     my $type = shift;
 
-    croak 'ref and type must be defined' unless defined $ref and defined $type;
+    croak _error 'ForceUsage' unless defined $ref and defined $type;
     bless \$ref, 'Bifcode::' . uc($type);
 }
 
@@ -280,8 +282,7 @@ sub _expand_bifcode {
 }
 
 sub diff_bifcode {
-    croak 'usage: diff_bifcode($b1, $b2, [$diff_args])'
-      unless @_ >= 2 and @_ <= 3;
+    croak _error 'DiffUsage' unless @_ >= 2 and @_ <= 3;
     my $b1        = shift;
     my $b2        = shift;
     my $diff_args = shift || { STYLE => 'Unified' };
@@ -307,14 +308,14 @@ Bifcode - simple serialization format
 
 =head1 VERSION
 
-0.001_10 (2017-11-06)
+0.001_11 (2017-11-08)
 
 
 =head1 SYNOPSIS
 
     use boolean;
     use Bifcode qw( encode_bifcode decode_bifcode );
- 
+
     my $bifcode = encode_bifcode {
         bools   => [ boolean::false, boolean::true, ],
         bytes   => \pack( 's<',       255 ),
@@ -334,14 +335,9 @@ Bifcode - simple serialization format
 
     my $decoded = decode_bifcode $bifcode;
 
-=head1 STATUS
-
-This module and related encoding format are still under development. Do
-not use it anywhere near production. Input is welcome.
-
 =head1 DESCRIPTION
 
-Bifcode implements the I<bifcode> serialisation format, a mixed
+B<Bifcode> implements the I<bifcode> serialisation format, a mixed
 binary/text encoding with support for the following data types:
 
 =over
@@ -381,7 +377,7 @@ is no need to escape special characters in strings. It is not
 considered human readable, but as it is mostly text it can usually be
 visually debugged.
 
-I<Bifcode> can only be constructed canonically; i.e. there is only one
+I<bifcode> can only be constructed canonically; i.e. there is only one
 possible encoding per data structure. This property makes it suitable
 for comparing structures (using cryptographic hashes) across networks.
 
@@ -391,27 +387,27 @@ with the same features.
 
 =head1 MOTIVATION & GOALS
 
-Bifcode was created for a project because none of currently available
-serialization formats (Bencode, JSON, MsgPack, Sereal, YAML, etc) met
-the requirements of:
+I<bifcode> was created for a project because none of currently
+available serialization formats (Bencode, JSON, MsgPack, Sereal, YAML,
+etc) met the requirements of:
 
 =over
 
 =item * Support for undef
 
-=item * Support for UTF8 strings
-
 =item * Support for binary data
 
-=item * Trivial to construct on the fly from within SQLite triggers
+=item * Support for UTF8 strings
 
 =item * Universally-recognized canonical form for hashing
 
+=item * Trivial to construct on the fly from within SQLite triggers
+
 =back
 
 I have no lofty goals or intentions to promote this outside of my
-specific case, but would appreciate hearing about other uses.
-Constructive discussion is welcome.
+specific case, but would appreciate hearing about other uses or
+implementations.  Constructive discussion is welcome.
 
 =head1 SPECIFICATION
 
@@ -550,87 +546,128 @@ Croaks if L<Text::Diff> is not installed.
 
 =head1 DIAGNOSTICS
 
+The following exceptions may be raised by B<Bifcode>:
+
 =over
 
-=item C<trailing garbage at %s>
+=item Bifcode::Error::Decode
 
-Your data does not end after the first I<bifcode>-serialised item.
+Your data is malformed in a non-identifiable way.
 
-You may also get this error if a malformed item follows.
+=item Bifcode::Error::DecodeBytes
 
-=item C<garbage at %s>
+Your data contains a byte string with an invalid length.
 
-Your data is malformed.
+=item Bifcode::Error::DecodeBytesTrunc
 
-=item C<unexpected end of data at %s>
+Your data includes a byte string declared to be longer than the
+available data.
 
-Your data is truncated.
-
-=item C<unexpected end of string data starting at %s>
-
-Your data includes a string declared to be longer than the available
-data.
-
-=item C<malformed string length at %s>
-
-Your data contained a string with negative length or a length with
-leading zeroes.
-
-=item C<malformed integer data at %s>
-
-Your data contained something that was supposed to be an integer but
-didn't make sense.
-
-=item C<dict key not in sort order at %s>
-
-Your data violates the I<bifcode> format constaint that dict keys must
-appear in lexical sort order.
-
-=item C<duplicate dict key at %s>
-
-Your data violates the I<bifcode> format constaint that all dict keys
-must be unique.
-
-=item C<dict key is not a string at %s>
-
-Your data violates the I<bifcode> format constaint that all dict keys
-be strings.
-
-=item C<dict key is missing value at %s>
-
-Your data contains a dictionary with an odd number of elements.
-
-=item C<nesting depth exceeded at %s>
+=item Bifcode::Error::DecodeDepth
 
 Your data contains dicts or lists that are nested deeper than the
 $max_depth passed to C<decode_bifcode()>.
 
-=item C<unhandled data type>
+=item Bifcode::Error::DecodeTrunc
 
-You are trying to serialise a data structure that consists of data
-types other than
+Your data is truncated.
 
-=over
+=item Bifcode::Error::DecodeFloat
 
-=item *
+Your data contained something that was supposed to be a float but
+didn't make sense.
 
-scalars
+=item Bifcode::Error::DecodeFloatTrunc
 
-=item *
+Your data contains a float that is truncated.
 
-references to arrays
+=item Bifcode::Error::DecodeInteger
 
-=item *
+Your data contained something that was supposed to be an integer but
+didn't make sense.
 
-references to hashes
+=item Bifcode::Error::DecodeIntegerTrunc
 
-=item *
+Your data contains an integer that is truncated.
 
-references to scalars
+=item Bifcode::Error::DecodeKeyType
 
-=back
+Your data violates the I<bifcode> format constaint that all dict keys
+be BIFCODE_BYTES or BIFCODE_UTF8.
 
-The format does not support this.
+=item Bifcode::Error::DecodeKeyDuplicate
+
+Your data violates the I<bifcode> format constaint that all dict keys
+must be unique.
+
+=item Bifcode::Error::DecodeKeyOrder
+
+Your data violates the I<bifcode> format constaint that dict keys must
+appear in lexical sort order.
+
+=item Bifcode::Error::DecodeKeyValue
+
+Your data contains a dictionary with an odd number of elements.
+
+=item Bifcode::Error::DecodeTrailing
+
+Your data does not end after the first I<bifcode>-serialised item.
+
+=item Bifcode::Error::DecodeUTF8
+
+Your data contained a UTF8 string with an invalid length.
+
+=item Bifcode::Error::DecodeUTF8Trunc
+
+Your data includes a string declared to be longer than the available
+data.
+
+=item Bifcode::Error::DecodeUsage
+
+You called C<decode_bifcode()> with invalid arguments.
+
+=item Bifcode::Error::DiffUsage
+
+You called C<diff_bifcode()> with invalid arguments.
+
+=item Bifcode::Error::EncodeBytesUndef
+
+You attempted to encode C<undef> as a byte string.
+
+=item Bifcode::Error::EncodeFloat
+
+You attempted to encode something as a float that isn't recognised as
+one.
+
+=item Bifcode::Error::EncodeFloatUndef
+
+You attempted to encode C<undef> as a float.
+
+=item Bifcode::Error::EncodeInteger
+
+You attempted to encode something as an integer that isn't recognised
+as one.
+
+=item Bifcode::Error::EncodeIntegerUndef
+
+You attempted to encode C<undef> as an integer.
+
+=item Bifcode::Error::EncodeUTF8Undef
+
+You attempted to encode C<undef> as a UTF8 string.
+
+=item Bifcode::Error::EncodeUnhandled
+
+You are trying to serialise a data structure that contains a data type
+not supported by the I<bifcode> format.
+
+=item Bifcode::Error::EncodeUsage
+
+You called C<encode_bifcode()> with invalid arguments.
+
+=item Bifcode::Error::ForceUsage
+
+You called C<force_bifcode()> with invalid arguments.
 
 =back
 
@@ -640,6 +677,10 @@ Strings and numbers are practically indistinguishable in Perl, so
 C<encode_bifcode()> has to resort to a heuristic to decide how to
 serialise a scalar. This cannot be fixed.
 
+At the moment all Perl hash keys are encoded as BIFCODE_UTF8 as I have
+not yet had the need for BIFCODE_BYTES keys or found a cheap, obvious
+way to distinguish the two.
+
 =head1 SEE ALSO
 
 This distribution includes the L<diff-bifcode> command-line utility for
